# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_TEP.ipynb.

# %% auto 0
__all__ = ['GetTEP']

# %% ../nbs/02_TEP.ipynb 3
import pandas as pd
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt

# %% ../nbs/02_TEP.ipynb 10
def GetTEP():
    filepath = 'https://github.com/waterboy96/TEPData/blob/1ac08a54cb9d420ff4bc0c3f0076ca06dc2ec7e4/TEP.csv?raw=true'
    TEP = pd.read_csv(filepath, index_col = [0])
    namepath = 'https://github.com/waterboy96/TEPData/blob/main/TEPnames.xlsx?raw=true'
    names = pd.read_excel(namepath, index_col = [0], header = None)
    TEP.columns = names.loc['Name']
    
    return TEP
    
    
