# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/dynamicPCA',
                'doc_host': 'https://waterboy96.github.io',
                'git_url': 'https://github.com/waterboy96/dynamicPCA',
                'lib_path': 'dPCA'},
  'syms': { 'dPCA.TEP': {'dPCA.TEP.GetTEP': ('tep.html#gettep', 'dPCA/TEP.py')},
            'dPCA.analysis': { 'dPCA.analysis.ACM_analysis': ('analysis.html#acm_analysis', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis': ('analysis.html#analysis', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis.__init__': ('analysis.html#analysis.__init__', 'dPCA/analysis.py'),
                               'dPCA.analysis.Analysis.show_plots': ('analysis.html#analysis.show_plots', 'dPCA/analysis.py'),
                               'dPCA.analysis.PACM_analysis': ('analysis.html#pacm_analysis', 'dPCA/analysis.py')},
            'dPCA.corrmat': { 'dPCA.corrmat.ACM': ('correlation_matrices.html#acm', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.PACM': ('correlation_matrices.html#pacm', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.correlation_matrix': ('correlation_matrices.html#correlation_matrix', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.imposter_matrix': ('correlation_matrices.html#imposter_matrix', 'dPCA/corrmat.py'),
                              'dPCA.corrmat.partial_correlation_matrix': ( 'correlation_matrices.html#partial_correlation_matrix',
                                                                           'dPCA/corrmat.py')},
            'dPCA.lag': { 'dPCA.lag.lag': ('lagfunctions.html#lag', 'dPCA/lag.py'),
                          'dPCA.lag.lag_uniform': ('lagfunctions.html#lag_uniform', 'dPCA/lag.py')}}}
